/*
Voxel terrain shader
Marijn Kentie 2009

changes over terrain:
-Guard against GPU hangs by aborting at max iterations
-Binary search for better intersection
-Rays don't start outside maximum terrain height
-Crater texture for deformations
-Tracing is ended when maximum length ray has been checked as opposed to when outside bounding box
-Dynamic choice of sampling rate
-Terrain occludes other objects using depth buffer
*/

#include "header.fxh"
#include "skybox.fx"
#include "smallbox.fx"

cbuffer constants
{	
	float HEIGHT_SCALE = 0.25f; //Heightmap height scaling
	float TEXTURE_REPEATS = 1.0f; //Tiling of the diffuse texture
	bool BINARY_SEARCH = true; //Run binary search for extra precision
	int MAX_BIN_ITERATIONS = 16; //Maximum iterations of bin. search
	int MAX_SAMPLES = 256; //Maximum amount of samples taken by ray tracing
	int MIN_SAMPLES = 128; //Minimum amount of samples
};

Texture2D textureCraters;

//--------------------------------------------------------------------------------------
// Vertex Shader
//--------------------------------------------------------------------------------------
PS_INPUT VS( VS_INPUT input )
{
	float BOX_SIZE=1.0;
	PS_INPUT output = (PS_INPUT)0;
	
	//Object to world space (no world transform needed as unit cube)
	output.pos=input.pos;
	
	//Set ray begin and endpoint
	output.eye = Eye;
	output.origPos = input.pos;
	
	//Final fixed function style transforms
	output.pos = mul( output.pos, View );
	output.pos = mul( output.pos, Projection );
	output.color = input.color;

	return output;
}



//--------------------------------------------------------------------------------------
// Pixel Shader
//--------------------------------------------------------------------------------------

float4 getHeight( float2 tex )
{
	//Sample heightmap(s)
	return (textureHeightMap.SampleLevel( samLinearClamp, tex,0 )-textureCraters.SampleLevel( samLinearClamp, tex,0 ))*HEIGHT_SCALE;
}

float2 binarySearch(float3 rayPos, float3 step)
{
	//Perform a binary search for heightmap intersection between ray position and previous ray position (rayPos-step)
	float middle = 0.5f;
	float offset = 0.25f;
	float2 coords;
	for(int i=0;i<MAX_BIN_ITERATIONS;i++)
	{
		coords = lerp(rayPos.xz,rayPos.xz-step.xz,middle);
		float height = getHeight(coords); 
		if(height > rayPos.y) //went out of volume
		{
			middle += offset;
		}
		else  if(height < rayPos.y) //went deeper into volume
		{
			middle -= offset;
		}
			
		offset /=2;
	}
	return coords ;
}

PS_OUTPUT PS( PS_INPUT input)
{	
	PS_OUTPUT o;
	float3 rayStart=input.eye;
	float3 rayDir = input.origPos-rayStart;

	//If ray starts outside of maximum terrain height, move it to the max height
	if(rayStart.y>HEIGHT_SCALE)
	{
		rayStart+=rayDir*((rayStart.y-HEIGHT_SCALE)/-rayDir.y);
		
		//If moved ray is outside bounding box, drop it
		if( rayStart.x<0 || rayStart.x>1 || rayStart.z<0 || rayStart.z>1)
		{
			discard;
		}
		rayDir = input.origPos-rayStart;
	}

	float rayLength=length(rayDir);
	rayDir/=rayLength;
	
	
	//Calculate step size by using formula to increase sampling rate along steep angles
	//  int nNumSteps = (int) lerp( g_nMaxSamples, g_nMinSamples, dot( vViewWS, vNormalWS ) ); from ATI paper
	int numSteps = (int) lerp(MIN_SAMPLES,MAX_SAMPLES,1-abs(rayDir.y));
	float stepSize=rayLength/numSteps;
	float3 step = rayDir*stepSize;
	float3 rayPos=rayStart;
	
	//Actual tracing
	for(int i=0;i<numSteps;i++)
	{	
		//Get terrain height; stepsize is added so height zero terrain is still drawn.
		if(getHeight(rayPos.xz).r+stepSize>rayPos.y)
			break;
		rayPos+=step;
	}
	
	//Check result of tracing
	if(i>=numSteps) //More samples were tried to be taken than should be possible
	{
		discard;
	}
	else //Succesful intersection
	{
		if(BINARY_SEARCH) //If binary searching is enabled, find a more precise intersection
		{
			rayPos.xz = binarySearch(rayPos,step);
		}
		o.color = textureDiffuse.SampleLevel(samLinear,rayPos.xz*TEXTURE_REPEATS,0);
	}
	
	//Assign depth to intersection point by transforming the ray to view*projection space
	float4 d = mul(float4(rayPos,1),View);
	d=mul(d,Projection);
	o.depth = d.z/d.w;
	return o;
}


//--------------------------------------------------------------------------------------
technique10 Render
{
	pass Skybox
	{
		SetVertexShader( CompileShader( vs_4_0, VS_sky() ) );
		SetGeometryShader( NULL );
		SetPixelShader( CompileShader( ps_4_0, PS_sky() ) );
		SetRasterizerState(Default);
	}

	pass Terrain
	{
	
		SetVertexShader( CompileShader( vs_4_0, VS() ) );
		SetGeometryShader( NULL );
		SetPixelShader( CompileShader( ps_4_0, PS() ) ); 
		SetDepthStencilState( EnableDepth, 0 );
		SetRasterizerState(Default);	
	}
	
   pass Smallbox
	{
		SetVertexShader( CompileShader( vs_4_0, VS_smallbox() ) );
		SetGeometryShader( NULL );
		SetPixelShader( CompileShader( ps_4_0, PS_smallbox() ) );  
		SetRasterizerState(RasSmallbox );
		SetDepthStencilState( EnableDepth, 0 );	
	} 
}