#include "d3d.h"
#include "dxhdr/d3d10.h"
#include "dxhdr/d3dx10.h"
#include <stdio.h>
#include <time.h>
#include <stdlib.h>

#include "main.h"
#include "window.h"

#define NUM_VERTS 36
const wchar_t* FX_FILE = L"terrain3.fx";

static ID3D10Device* device;
static IDXGISwapChain* swapChain;
static ID3D10RenderTargetView* renderTargetView;
static ID3D10Effect* effect;
static ID3D10EffectTechnique* technique;
static ID3D10InputLayout* vertexLayout;
static ID3D10Buffer* vertexBuffer;


struct
{
ID3D10Texture2D* texture;
ID3D10DepthStencilView* view;
}
depthStencil;

//Texture related variables
struct Texture
{
ID3D10EffectShaderResourceVariable* variable;
ID3D10ShaderResourceView* view;
};

static struct
{
	Texture diffuse;
	Texture heightmap;
	Texture sky;
	Texture craters;
} textures;

//Shader-side transformation matrices
static struct
{
	ID3D10EffectMatrixVariable* world;
	ID3D10EffectMatrixVariable* view;
	ID3D10EffectMatrixVariable* projection;
	ID3D10EffectVectorVariable* eye;

}
transformsShader;

//CPU-side transformation matrices
static struct
{
	D3DXMATRIX world;
	D3DXMATRIX view;
	D3DXMATRIX projection;
	D3DXVECTOR3 eye;
}
transforms;

//Misc shader vars for settings toggles
ID3D10EffectScalarVariable* shaderVars[6];

//Vertex format
struct Vertex
{
	D3DXVECTOR3 Pos;
	D3DXCOLOR Color;
	D3DXVECTOR3 Normal;
	D3DXVECTOR2 TexCoord;
};



static bool setupD3D()
{
	HRESULT hr; //used to check function return values

	//Create device and swap chain
	DXGI_SWAP_CHAIN_DESC sd;
	ZeroMemory( &sd, sizeof( sd ) );
	sd.BufferCount = 1;
	sd.BufferDesc.Width = Window::getWidth();
	sd.BufferDesc.Height = Window::getHeight();
	sd.BufferDesc.Format = DXGI_FORMAT_R8G8B8A8_UNORM;
	sd.BufferDesc.RefreshRate.Numerator = 60;
	sd.BufferDesc.RefreshRate.Denominator = 1;
	sd.BufferUsage = DXGI_USAGE_RENDER_TARGET_OUTPUT;
	sd.OutputWindow = Window::getHWND();
	sd.SampleDesc.Count = 1;
	sd.SampleDesc.Quality = 0;
	sd.Windowed = TRUE;

	UINT flags = 0;
#ifdef _DEBUG
	flags = D3D10_CREATE_DEVICE_DEBUG;
#endif 
	hr = D3D10CreateDeviceAndSwapChain( NULL, D3D10_DRIVER_TYPE_HARDWARE, NULL, flags, D3D10_SDK_VERSION, &sd, &swapChain, &device );
	if(FAILED(hr))
	{
		puts("Error creating device/swap chain.");
		return false;
	}

	//Create a render target view
	ID3D10Texture2D* pBuffer;
	hr = swapChain->GetBuffer( 0, __uuidof( ID3D10Texture2D ), ( LPVOID* )&pBuffer );
	if(FAILED(hr))
	{
		puts("Error getting swap chain buffer.");
		return false;
	}

	hr = device->CreateRenderTargetView( pBuffer, NULL, &renderTargetView );
	pBuffer->Release();
	if( FAILED( hr ) )
	{
		puts("Error creating render target view.");
		return false;
	}
	device->OMSetRenderTargets( 1, &renderTargetView, NULL );

	// Setup the viewport
	D3D10_VIEWPORT vp;
	vp.Width = Window::getWidth();
	vp.Height = Window::getHeight();
	vp.MinDepth = 0.0f;
	vp.MaxDepth = 1.0f;
	vp.TopLeftX = 0;
	vp.TopLeftY = 0;
	device->RSSetViewports( 1, &vp );

	return true;
}

static bool setupDepthBuffer()
{
	//Create depth stencil texture
	D3D10_TEXTURE2D_DESC descDepth;
	DXGI_SWAP_CHAIN_DESC scd;
	swapChain->GetDesc(&scd);
	descDepth.Width = scd.BufferDesc.Width;
	descDepth.Height = scd.BufferDesc.Height;
	descDepth.MipLevels = 1;
	descDepth.ArraySize = 1;
	descDepth.Format =  DXGI_FORMAT_R32_TYPELESS ;
	descDepth.SampleDesc.Count = 1;
	descDepth.SampleDesc.Quality = 0;
	descDepth.Usage = D3D10_USAGE_DEFAULT;
	descDepth.BindFlags = D3D10_BIND_DEPTH_STENCIL;
	descDepth.CPUAccessFlags = 0;
	descDepth.MiscFlags = 0;
	if(FAILED(device->CreateTexture2D( &descDepth, NULL, &depthStencil.texture )))
	{
		puts("Depth texture creation failed.");
		return false;
	}

	// Create the depth stencil view
	D3D10_DEPTH_STENCIL_VIEW_DESC descDSV;
	descDSV.Format = DXGI_FORMAT_D32_FLOAT;
	descDSV.ViewDimension = D3D10_DSV_DIMENSION_TEXTURE2D;
	descDSV.Texture2D.MipSlice=0;
	
	if(FAILED(device->CreateDepthStencilView( depthStencil.texture,&descDSV,&depthStencil.view )))
	{
		puts("Depth stencil view creation failed.");
		return false;
	}

	// Bind the depth stencil view
	device->OMSetRenderTargets(1,&renderTargetView,depthStencil.view);

	return true;	
}

static bool loadEffects()
{
	ID3D10Blob* blob;	
		
	//Create the effect we'll be using
	DWORD dwShaderFlags = D3D10_SHADER_ENABLE_STRICTNESS;
#ifdef _DEBUG
    dwShaderFlags |= D3D10_SHADER_DEBUG;
#endif
	bool b;
	b=FAILED(D3DX10CreateEffectFromFile(FX_FILE,NULL,NULL,"fx_4_0",dwShaderFlags,NULL,device,NULL,NULL,&effect,&blob,NULL));
	if(blob) //Show compile errors if present
			puts((char*) blob->GetBufferPointer());		
	if(b)
	{
		puts("Error compiling effect.");		
		return false;
	}
	
	
	//Get technique from effect
	if((technique=effect->GetTechniqueByName("Render"))==NULL)
	{
		puts("Technique not found.");
		return false;
	}

	shaderVars[D3D::BINARY_SEARCH] = effect->GetVariableByName("BINARY_SEARCH")->AsScalar();
	shaderVars[D3D::LIGHTING] = effect->GetVariableByName("LIGHTING")->AsScalar();
	shaderVars[D3D::OCCLUSION] = effect->GetVariableByName("OCCLUSION")->AsScalar();
	shaderVars[D3D::TERRAIN] = effect->GetVariableByName("TERRAIN")->AsScalar();
	shaderVars[D3D::DEFORMATION] = effect->GetVariableByName("DEFORMATION")->AsScalar();
	shaderVars[D3D::DYNAMIC_SAMPLE] = effect->GetVariableByName("DYNAMIC_SAMPLE")->AsScalar();
	
   return true;
}

static bool setInputLayout()
{
	// Define the input layout
    D3D10_INPUT_ELEMENT_DESC layoutDesc[] =
    {
		{ "POSITION",   0, DXGI_FORMAT_R32G32B32_FLOAT, 0, D3D10_APPEND_ALIGNED_ELEMENT,   D3D10_INPUT_PER_VERTEX_DATA, 0 },
		{ "COLOR",   0, DXGI_FORMAT_R32G32B32A32_FLOAT, 0, D3D10_APPEND_ALIGNED_ELEMENT,   D3D10_INPUT_PER_VERTEX_DATA, 0 },
        { "NORMAL",     0, DXGI_FORMAT_R32G32B32_FLOAT, 0, D3D10_APPEND_ALIGNED_ELEMENT,  D3D10_INPUT_PER_VERTEX_DATA, 0 },
		{ "TEXCOORD",     0, DXGI_FORMAT_R32G32_FLOAT, 0, D3D10_APPEND_ALIGNED_ELEMENT,  D3D10_INPUT_PER_VERTEX_DATA, 0 },
    };
    UINT numElements = sizeof( layoutDesc ) / sizeof( layoutDesc[0] );

    // Create the input layout
    D3D10_PASS_DESC passDesc;		
    technique->GetPassByIndex( 0 )->GetDesc(&passDesc);
    if(FAILED(device->CreateInputLayout(layoutDesc, numElements, passDesc.pIAInputSignature, passDesc.IAInputSignatureSize, &vertexLayout )))
	{
		puts("Error creating input layout.");
		return false;
	}
	device->IASetInputLayout(vertexLayout);
	return true;	
}

static bool createVertexBuffers()
{
	const float yscale=0.5f;
	//Define initial vertex buffer
	Vertex vertices[] =
    {
		
		//left
		{ D3DXVECTOR3( 0, 0, 0  ),D3DXCOLOR(0.0f,0.0f,0.0f,0.0f),D3DXVECTOR3( 0.0f, 0.0f, 0.0f ),D3DXVECTOR2(0.0f, 0.0f)},
		{ D3DXVECTOR3( 0, 1, 0 ),D3DXCOLOR(0.0f,0.0f,0.0f,0.0f),D3DXVECTOR3( 0.0f, 0.0f, 0.0f ),D3DXVECTOR2(0.0f, 0.0f)},
		{ D3DXVECTOR3( 0, 0, 1 ),D3DXCOLOR(0.0f,0.0f,0.0f,0.0f),D3DXVECTOR3( 0.0f, 0.0f, 0.0f ),D3DXVECTOR2(0.0f, 0.0f)},
		{ D3DXVECTOR3(  0, 0, 1 ),D3DXCOLOR(0.0f,0.0f,0.0f,0.0f),D3DXVECTOR3( 0.0f, 0.0f, 0.0f ),D3DXVECTOR2(0.0f, 0.0f)},
		{ D3DXVECTOR3( 0, 1, 0 ),D3DXCOLOR(0.0f,0.0f,0.0f,0.0f),D3DXVECTOR3( 0.0f, 0.0f, 0.0f ),D3DXVECTOR2(0.0f, 0.0f)},
		{ D3DXVECTOR3( 0, 1, 1  ),D3DXCOLOR(0.0f,0.0f,0.0f,0.0f),D3DXVECTOR3( 0.0f, 0.0f, 0.0f ),D3DXVECTOR2(0.0f, 0.0f)},
		//right
		{ D3DXVECTOR3( 1, 0, 1 ),D3DXCOLOR(0.0f,0.0f,0.0f,0.0f),D3DXVECTOR3( 0.0f, 0.0f, 0.0f ),D3DXVECTOR2(0.0f, 0.0f)},
		{ D3DXVECTOR3( 1, 1, 1 ),D3DXCOLOR(0.0f,0.0f,0.0f,0.0f),D3DXVECTOR3( 0.0f, 0.0f, 0.0f ),D3DXVECTOR2(0.0f, 0.0f)},
		{ D3DXVECTOR3( 1, 0, 0 ),D3DXCOLOR(0.0f,0.0f,0.0f,0.0f),D3DXVECTOR3( 0.0f, 0.0f, 0.0f ),D3DXVECTOR2(0.0f, 0.0f)},
		{ D3DXVECTOR3( 1, 0, 0 ),D3DXCOLOR(0.0f,0.0f,0.0f,0.0f),D3DXVECTOR3( 0.0f, 0.0f, 0.0f ),D3DXVECTOR2(0.0f, 0.0f)},
		{ D3DXVECTOR3( 1, 1, 1 ),D3DXCOLOR(0.0f,0.0f,0.0f,0.0f),D3DXVECTOR3( 0.0f, 0.0f, 0.0f ),D3DXVECTOR2(0.0f, 0.0f)},
		{ D3DXVECTOR3( 1, 1, 0 ),D3DXCOLOR(0.0f,0.0f,0.0f,0.0f),D3DXVECTOR3( 0.0f, 0.0f, 0.0f ),D3DXVECTOR2(0.0f, 0.0f)},
		//back
		{ D3DXVECTOR3( 0, 0, 1),D3DXCOLOR(0.0f,0.0f,0.0f,0.0f),D3DXVECTOR3( 0.0f, 0.0f, 0.0f ),D3DXVECTOR2(0.0f, 0.0f)},
		{ D3DXVECTOR3( 0, 1, 1 ),D3DXCOLOR(0.0f,0.0f,0.0f,0.0f),D3DXVECTOR3( 0.0f, 0.0f, 0.0f ),D3DXVECTOR2(0.0f, 0.0f)},
		{ D3DXVECTOR3(1, 0, 1 ),D3DXCOLOR(0.0f,0.0f,0.0f,0.0f),D3DXVECTOR3( 0.0f, 0.0f, 0.0f ),D3DXVECTOR2(0.0f, 0.0f)},
		{ D3DXVECTOR3( 1, 0, 1 ),D3DXCOLOR(0.0f,0.0f,0.0f,0.0f),D3DXVECTOR3( 0.0f, 0.0f, 0.0f ),D3DXVECTOR2(0.0f, 0.0f)},
		{ D3DXVECTOR3( 0, 1, 1  ),D3DXCOLOR(0.0f,0.0f,0.0f,0.0f),D3DXVECTOR3( 0.0f, 0.0f, 0.0f ),D3DXVECTOR2(0.0f, 0.0f)},
		{ D3DXVECTOR3( 1, 1, 1 ),D3DXCOLOR(0.0f,0.0f,0.0f,0.0f),D3DXVECTOR3( 0.0f, 0.0f, 0.0f ),D3DXVECTOR2(0.0f, 0.0f)},
		//front
		{ D3DXVECTOR3( 1, 0, 0  ),D3DXCOLOR(0.0f,0.0f,0.0f,0.0f),D3DXVECTOR3( 0.0f, 0.0f, 0.0f ),D3DXVECTOR2(0.0f, 0.0f)},
		{ D3DXVECTOR3( 1, 1, 0 ),D3DXCOLOR(0.0f,0.0f,0.0f,0.0f),D3DXVECTOR3( 0.0f, 0.0f, 0.0f ),D3DXVECTOR2(0.0f, 0.0f)},
		{ D3DXVECTOR3( 0, 0, 0  ),D3DXCOLOR(0.0f,0.0f,0.0f,0.0f),D3DXVECTOR3( 0.0f, 0.0f, 0.0f ),D3DXVECTOR2(0.0f, 0.0f)},
		{ D3DXVECTOR3( 0, 0, 0  ),D3DXCOLOR(0.0f,0.0f,0.0f,0.0f),D3DXVECTOR3( 0.0f, 0.0f, 0.0f ),D3DXVECTOR2(0.0f, 0.0f)},
		{ D3DXVECTOR3( 1, 1, 0  ),D3DXCOLOR(0.0f,0.0f,0.0f,0.0f),D3DXVECTOR3( 0.0f, 0.0f, 0.0f ),D3DXVECTOR2(0.0f, 0.0f)},
		{ D3DXVECTOR3( 0, 1, 0  ),D3DXCOLOR(0.0f,0.0f,0.0f,0.0f),D3DXVECTOR3( 0.0f, 0.0f, 0.0f ),D3DXVECTOR2(0.0f, 0.0f)},		
		//bottom
		{ D3DXVECTOR3( 0,0,0 ),D3DXCOLOR(0.0f,0.0f,0.0f,0.0f),D3DXVECTOR3( 0.0f, 0.0f, 0.0f ),D3DXVECTOR2(0.0f, 0.0f)},
		{ D3DXVECTOR3(0,0,1 ),D3DXCOLOR(0.0f,0.0f,0.0f,0.0f),D3DXVECTOR3( 0.0f, 0.0f, 0.0f ),D3DXVECTOR2(0.0f, 0.0f)},
		{ D3DXVECTOR3( 1, 0, 0 ),D3DXCOLOR(0.0f,0.0f,0.0f,0.0f),D3DXVECTOR3( 0.0f, 0.0f, 0.0f ),D3DXVECTOR2(0.0f, 0.0f)},
		{ D3DXVECTOR3(  1, 0, 0  ),D3DXCOLOR(0.0f,0.0f,0.0f,0.0f),D3DXVECTOR3( 0.0f, 0.0f, 0.0f ),D3DXVECTOR2(0.0f, 0.0f)},
		{ D3DXVECTOR3( 0, 0, 1),D3DXCOLOR(0.0f,0.0f,0.0f,0.0f),D3DXVECTOR3( 0.0f, 0.0f, 0.0f ),D3DXVECTOR2(0.0f, 0.0f)},
		{ D3DXVECTOR3( 1, 0, 1  ),D3DXCOLOR(0.0f,0.0f,0.0f,0.0f),D3DXVECTOR3( 0.0f, 0.0f, 0.0f ),D3DXVECTOR2(0.0f, 0.0f)},	
		//top
		{ D3DXVECTOR3( 1, 1, 1),D3DXCOLOR(0.0f,0.0f,0.0f,0.0f),D3DXVECTOR3( 0.0f, 0.0f, 0.0f ),D3DXVECTOR2(0.0f, 0.0f)},
		{ D3DXVECTOR3( 0, 1, 1 ),D3DXCOLOR(0.0f,0.0f,0.0f,0.0f),D3DXVECTOR3( 0.0f, 0.0f, 0.0f ),D3DXVECTOR2(0.0f, 0.0f)},
		{ D3DXVECTOR3( 0, 1, 0  ),D3DXCOLOR(0.0f,0.0f,0.0f,0.0f),D3DXVECTOR3( 0.0f, 0.0f, 0.0f ),D3DXVECTOR2(0.0f, 0.0f)},
		{ D3DXVECTOR3( 1, 1, 1  ),D3DXCOLOR(0.0f,0.0f,0.0f,0.0f),D3DXVECTOR3( 0.0f, 0.0f, 0.0f ),D3DXVECTOR2(0.0f, 0.0f)},
		{ D3DXVECTOR3( 0, 1, 0 ),D3DXCOLOR(0.0f,0.0f,0.0f,0.0f),D3DXVECTOR3( 0.0f, 0.0f, 0.0f ),D3DXVECTOR2(0.0f, 0.0f)},		
		{ D3DXVECTOR3( 1, 1, 0),D3DXCOLOR(0.0f,0.0f,0.0f,0.0f),D3DXVECTOR3( 0.0f, 0.0f, 0.0f ),D3DXVECTOR2(0.0f, 0.0f)},
    };


	D3D10_SUBRESOURCE_DATA initialData;
	ZeroMemory(&initialData,sizeof(initialData));
    initialData.pSysMem = vertices;


	//Set up vertex buffer
	D3D10_BUFFER_DESC bufferDesc;
    bufferDesc.Usage            = D3D10_USAGE_DEFAULT;
    bufferDesc.ByteWidth        = sizeof( Vertex ) *NUM_VERTS;
    bufferDesc.BindFlags        = D3D10_BIND_VERTEX_BUFFER;
    bufferDesc.CPUAccessFlags   = 0;
    bufferDesc.MiscFlags        = 0;
	
    if(FAILED(device->CreateBuffer( &bufferDesc,&initialData, &vertexBuffer )))
	{
		puts("Error creating vertex buffer.");
		return false;
	}

	UINT stride=sizeof(Vertex);
	UINT offset = 0;
	device->IASetVertexBuffers( 0, 1, &vertexBuffer, &stride, &offset );

	//Set primitive topology
	device->IASetPrimitiveTopology( D3D10_PRIMITIVE_TOPOLOGY_TRIANGLELIST );
	return true;
}


void D3D::setViewMatrix(D3DXMATRIX &view,D3DXVECTOR3& eye)
{
	//Also requires setting eye point as the two are closely related and should be in sync
	transforms.view = view;
	transforms.eye = eye;
	
	
}

void D3D::setProjectionMatrix(int width,int height)
{
	D3DXMatrixPerspectiveFovLH( &transforms.projection, ( float )D3DX_PI * 0.25f, width / ( FLOAT )height, 0.001f, 500.0f );	
	transformsShader.projection->SetMatrix((float*)&transforms.projection);
}

static bool initMatrices()
{
	//Bind transformation matrices to shaders and initialize them

	transformsShader.world = effect->GetVariableByName("World")->AsMatrix();
  	transformsShader.view = effect->GetVariableByName("View")->AsMatrix();
	transformsShader.projection = effect->GetVariableByName("Projection")->AsMatrix();
	transformsShader.eye = effect->GetVariableByName("Eye")->AsVector();
	
	D3DXMatrixIdentity(&transforms.world);
	D3DXMatrixIdentity(&transforms.view);	
	D3D::setProjectionMatrix(Window::getWidth(),Window::getHeight());
	return true;	
}






static bool loadTexture(Texture &texture,char* name,D3DX10_IMAGE_LOAD_INFO* info)
{
	texture.variable = effect->GetVariableByName(name)->AsShaderResource();

	char filename[MAX_PATH];
	strcpy(filename,name);
	strcat(filename,".dds");

    if( FAILED(D3DX10CreateShaderResourceViewFromFileA( device, filename, info, NULL, &texture.view, NULL ) ) )
     {
		puts("Error loading texture.");
		return false;
	}	
	texture.variable->SetResource(texture.view);
	return true;
}

static bool loadTextures()
{
	return loadTexture(textures.diffuse,"textureDiffuse",NULL) &&
	loadTexture(textures.heightmap,"textureHeightMap",NULL) &&
	loadTexture(textures.sky,"textureSkybox",NULL) &&
	loadTexture(textures.craters,"textureCraters",NULL)	;
}

bool D3D::init()
{
	puts("Initializing Direct3D.");
	return setupD3D()  && loadEffects() && setInputLayout() && createVertexBuffers() && setupDepthBuffer() && initMatrices() && loadTextures();
}

void D3D::cleanup()
{
	if(device)
		device->ClearState();
	
	SAFE_RELEASE(vertexBuffer);
	SAFE_RELEASE(vertexLayout);
	SAFE_RELEASE(swapChain);
	SAFE_RELEASE(renderTargetView);
	SAFE_RELEASE(depthStencil.texture);
	SAFE_RELEASE(depthStencil.view);
	SAFE_RELEASE(effect);
	SAFE_RELEASE(device);
	SAFE_RELEASE(textures.diffuse.view);
	SAFE_RELEASE(textures.heightmap.view);
	SAFE_RELEASE(textures.sky.view);
	SAFE_RELEASE(textures.craters.view);
}

void D3D::render()
{ 
	
	
	//Clear screen
	float clearColor[4] = { 0.0f, 0.125f, 0.3f, 1.0f };
	device->ClearRenderTargetView(renderTargetView, clearColor );
	device->ClearDepthStencilView(depthStencil.view, D3D10_CLEAR_DEPTH, 1.0, 0 );
   
	 
	//Update shader-side transform matrices
	transformsShader.world->SetMatrix((float*) &transforms.world);
	transformsShader.view->SetMatrix((float*) &transforms.view);
	transformsShader.eye->SetFloatVector((float*) &transforms.eye);

	//Render vertex data
	D3D10_TECHNIQUE_DESC techDesc;
    technique->GetDesc(&techDesc);
    for(UINT p = 0; p < techDesc.Passes; ++p )
    {
		technique->GetPassByIndex( p )->Apply( 0 );		
		device->Draw(NUM_VERTS,0);
    }
	swapChain->Present(0,0);
	
}

void D3D::addCraters()
{
	//Load crater texture
	ID3D10Resource* src = NULL;
	D3DX10CreateTextureFromFile( device, L"textureCrater.dds", NULL, NULL, &src, NULL );
	ID3D10Texture2D* tex2Dsrc;
	src->QueryInterface(__uuidof(ID3D10Texture2D),(void**) &tex2Dsrc);
	D3D10_TEXTURE2D_DESC descsrc;
	tex2Dsrc->GetDesc(&descsrc);

	//Get 'cratermap'
	ID3D10Resource* dest;
	textures.craters.view->GetResource(&dest);	
	ID3D10Texture2D* tex2Ddest;
	dest->QueryInterface(__uuidof(ID3D10Texture2D),(void**) &tex2Ddest);
	D3D10_TEXTURE2D_DESC desc;
	tex2Ddest->GetDesc(&desc);
	
	srand(time(NULL));

	//Draw the craters to the cratermap in five random places
	for(int i=0;i<5;i++)
	{
		int h = rand()%(desc.Height-descsrc.Height);
		int w = rand()%(desc.Width-descsrc.Width);
		device->CopySubresourceRegion(dest,0,h,w,0,src,0,NULL);
	}
		
	SAFE_RELEASE(src);
	SAFE_RELEASE(dest);
	SAFE_RELEASE(tex2Ddest);
	SAFE_RELEASE(tex2Dsrc);
}


void D3D::toggleSetting(int setting)
{
	BOOL b;
	shaderVars[setting]->GetBool(&b);
	b=!b;
	shaderVars[setting]->SetBool(b);
	printf("Toggled: %d\n",b);

}